"""Annotate family
"""
from ..i18n import _
from ..error import DictConsistencyError
from ..utils import normalize_family


#mode order is important
modes_level = ('basic', 'normal', 'expert')


class Mode:  # pylint: disable=R0903
    """Class to manage mode level
    """
    def __init__(self,
                 name: str,
                 level: int,
                 ) -> None:
        self.name = name
        self.level = level

    def __gt__(self,
               other: int,
               ) -> bool:
        return other.level < self.level


modes = {name: Mode(name, idx) for idx, name in enumerate(modes_level)}


class FamilyAnnotator:
    """Annotate family
    """
    def __init__(self,
                 objectspace,
                 ):
        self.objectspace = objectspace
        if not hasattr(self.objectspace.space, 'variables'):
            return
        self.remove_empty_families()
        self.family_names()
        self.change_modes()
        self.dynamic_families()
        self.convert_help()

    def remove_empty_families(self) -> None:
        """Remove all families without any variable
        """
        for families in self.objectspace.space.variables.values():
            removed_families = []
            for family_name, family in families.family.items():
                if not hasattr(family, 'variable') or len(family.variable) == 0:
                    removed_families.append(family_name)
            for family_name in removed_families:
                del families.family[family_name]

    def family_names(self) -> None:
        """Set doc, path, ... to family
        """
        for families in self.objectspace.space.variables.values():
            families.doc = families.name
            families.path = families.name
            for family in families.family.values():
                if not hasattr(family, 'description'):
                    family.description = family.name
                family.doc = family.description
                del family.description
                family.name = normalize_family(family.name)

    def change_modes(self):
        """change the mode of variables
        """
        for families in self.objectspace.space.variables.values():
            for family in families.family.values():
                family_mode = family.mode
                # default is high level
                new_family_mode = modes_level[-1]
                for variable in family.variable.values():
                    if not isinstance(variable, self.objectspace.leadership):
                        func = self._change_variabe_mode
                    else:
                        func = self._change_variable_mode_leader
                    func(variable,
                         family_mode,
                         )
                    if modes[new_family_mode] > modes[variable.mode]:
                        new_family_mode = variable.mode
                # set the lower variable mode to family
                family.mode = new_family_mode

    def _change_variabe_mode(self,
                             variable,
                             family_mode: str,
                             is_follower=False,
                             ) -> None:
        # auto_save or auto_freeze variable is set to 'basic' mode
        # if its mode is not defined by the user
        if 'mode' not in vars(variable) and \
                (variable.auto_save is True or variable.auto_freeze is True):
            variable.mode = modes_level[0]
        # mandatory variable without value is a basic variable
        if variable.mandatory is True and not hasattr(variable, 'default'):
            variable.mode = modes_level[0]
        # none basic variable in high level family has to be in high level
        if modes[variable.mode] < modes[family_mode] and \
                (not is_follower or variable.mode != modes_level[0]):
            variable.mode = family_mode

    def _change_variable_mode_leader(self,
                                     leadership,
                                     family_mode: str,
                                     ) -> None:
        is_follower = False
        leader_mode = None
        for follower in leadership.variable:
            if follower.auto_save is True:
                xmlfiles = self.objectspace.display_xmlfiles(leadership.xmlfiles)
                msg = _(f'leader/followers "{follower.name}" could not be auto_save in {xmlfiles}')
                raise DictConsistencyError(msg, 29)
            if follower.auto_freeze is True:
                xmlfiles = self.objectspace.display_xmlfiles(leadership.xmlfiles)
                msg = f'leader/followers "{follower.name}" could not be auto_freeze in {xmlfiles}'
                raise DictConsistencyError(_(msg), 30)
            self._change_variabe_mode(follower,
                                      family_mode,
                                      is_follower,
                                      )
            if leader_mode is None:
                leader_mode = leadership.variable[0].mode
                leadership.variable[0].mode = None
            else:
                # leader's mode is minimum level
                if modes[leader_mode] > modes[follower.mode]:
                    follower.mode = leader_mode
            is_follower = True
        leadership.mode = leader_mode

    def dynamic_families(self):
        """link dynamic families to object
        """
        for families in self.objectspace.space.variables.values():
            for family in families.family.values():
                if 'dynamic' not in vars(family):
                    continue
                family.suffixes = self.objectspace.paths.get_variable(family.dynamic)
                del family.dynamic
                if not family.suffixes.multi:
                    xmlfiles = self.objectspace.display_xmlfiles(family.xmlfiles)
                    msg = _(f'dynamic family "{family.name}" must be linked '
                            f'to multi variable in {xmlfiles}')
                    raise DictConsistencyError(msg, 16)

    def convert_help(self):
        """Convert variable help
        """
        for families in self.objectspace.space.variables.values():
            for family in families.family.values():
                if not hasattr(family, 'help'):
                    continue
                if not hasattr(family, 'information'):
                    family.information = self.objectspace.information(family.xmlfiles)
                family.information.help = family.help
                del family.help
